import java.util.Arrays;

public class Simplex {
    private boolean debugging;
    private Matrix matEcart;
    private Matrix tabAux;
    private int nbSousCondition;
    private int ligne; // Ligne
    private int colonne; // Colonne
    private int nbPivot;
    private static double EPSILON = 0E-7;
    private int nbContraintes;

    public Matrix getMatEcart() {
        return matEcart;
    }

    public int getNbPivot() {
        return nbPivot;
    }

    public Simplex(int ligne, int colonne, int nbSousCondition, int nbContraintes, boolean debugging) {
        this.ligne = ligne;
        this.colonne = colonne;
        this.matEcart = new Matrix(ligne, colonne);
        this.nbSousCondition = nbSousCondition;
        this.tabAux = new Matrix(ligne + 2, colonne + nbSousCondition + 1);
        this.nbPivot = 0;
        this.nbContraintes = nbContraintes;
        this.debugging = debugging;
    }

    void createSimplex(Equation eq, int nbContraintes) {
        // Matrice Amxn
        for (int i = 0; i < this.ligne; i++) {
            for (int j = 0; j < this.colonne; j++) {
                if (j < nbContraintes) {
                    this.matEcart.setData(i, j, eq.getMatAtId(i, j));
                } else {
                    this.matEcart.setData(i, j, j - nbContraintes == i ? 1.0 : 0.0);
                }
            }
        }

        // Membre de droite
        for (int i = 0; i <= this.ligne - 1; i++)
            this.matEcart.setData(i, this.colonne - 1, eq.getRightVec().get(i));

        // Fonction obj
        this.matEcart.matrixRealloc(this.matEcart.getLine() + 1, this.matEcart.getCol());
        for (int i = 0; i < nbContraintes; i++)
            this.matEcart.setData(this.ligne, i, eq.getFuncObj().get(i));
    }

    /**
     * Si b[i] < 0 phase 1 sinon phase 2
     *
     * @return true = phase 1 | false = phase 2
     */
    boolean which_phase() {
        for (int i = 0; i < this.ligne; i++) {
            if (signe(this.matEcart.getData(i, this.colonne - 1))) return true;
        }
        return false;
    }

    void tabAux() {
        if (debugging) System.out.println();
        double[] tabRes = new double[this.colonne + this.nbSousCondition + 1];
        Arrays.fill(tabRes, 1.0);
        for (int i = 0; i < this.colonne; i++) {
            for (int j = 0; j < this.ligne; j++) {
                if (this.matEcart.getData(j, this.colonne - 1) < 0) {
                    if (this.matEcart.getData(j, i) != 0) {
                        this.matEcart.setData(j, i, -this.matEcart.getData(j, i));
                    }
                }
            }
        }
        for (int j = 0; j < this.colonne; j++) {
            double tmpSum = 0;
            for (int i = 0; i < this.ligne; i++) {
                tmpSum += -1 * this.matEcart.getData(i, j);
            }
            tabRes[j] = tmpSum;
        }

        // -2 car => tabRes[(tabRes.length-1) - (this.nbSousCondition - 1)];
        tabRes[tabRes.length - 1] = tabRes[tabRes.length - this.nbSousCondition - 2];
        tabRes[tabRes.length - this.nbSousCondition - 2] = 0;
        for (int i = 0; i < this.tabAux.getLine(); i++) {
            for (int j = 0; j < this.tabAux.getCol(); j++) {
                if (i < this.matEcart.getLine() && j < this.matEcart.getCol()) {
                    this.tabAux.setData(i, j, this.matEcart.getData(i, j));
                } else if (i == this.tabAux.getLine() - 1) {
                    this.tabAux.setData(i, j, tabRes[j]);
                } else // membre de droite à la fin du tab
                {
                    this.tabAux.setData(i, j, this.matEcart.getData(i, j - this.matEcart.getLine()));
                }
            }
        }
        if(debugging) this.tabAux.printTabAux("Tableau auxiliaire", this.nbContraintes, this.nbSousCondition,
                this.tabAux.getCol() - this.matEcart.getCol() - 1);

        pivot(this.tabAux, true);
        double solutionOptimale = this.tabAux.getData(this.tabAux.getLine() - 1, this.tabAux.getCol() - 1);
        if (solutionOptimale > 0 + EPSILON) {
            System.out.println("Il n'y a pas de solutions admissibles pour ce problème.");
        }
        if (Math.abs(solutionOptimale) < EPSILON || solutionOptimale == 0) {
            // Il y a une solution optimale
            // Il faut enlever les variables auxilaires
            Matrix res = new Matrix(matEcart.getLine(), matEcart.getCol());
            res.matrixFill(res.getLine(), res.getCol(), tabAux.getDatas());
            if (debugging) res.matrixPrint("Petit tableau");
            nbPivot = 0;
            pivot(res, true);
            if (debugging) {
                res.matrixPrint("Matrice Résultat ");
                System.out.println("Nombre de pivot : " + nbPivot);
            }
        }
    }

    int getFirstNeg(Matrix mat) {
        for (int j = 0; j < mat.getCol() - 1; j++) {
            if (signe(mat.getData(mat.getLine() - 1, j))) return j;
        }
        return -1;
    }

    /**
     * @param mat   Matrix
     * @param phase true => phase 1 | false => phase 2
     */
    void pivot(Matrix mat, boolean phase) {
        this.nbPivot += 1;
        int firstNeg = getFirstNeg(mat);
        if (firstNeg == -1) return; // si pas de négatif
        boolean has_neg = false;
        int id = ligneSortante(mat, firstNeg, phase);
        double val_pivot = mat.getData(id, firstNeg);

        for (int i = 0; i < mat.getCol(); i++) {
            mat.setData(id, i, mat.getData(id, i) / val_pivot);
        }

        for (int i = 0; i < mat.getLine(); i++) {
            val_pivot = mat.getData(i, firstNeg);
            for (int j = 0; j < mat.getCol(); j++) {
                if (i != id) {
                    mat.setData(i, j, mat.getData(i, j) - val_pivot * mat.getData(id, j));
                }
            }
        }
        if (debugging) {
            mat.matrixPrint("Pivot numéro " + this.nbPivot);
            System.out.println("colonne du pivot: " + firstNeg);
            System.out.println("ligne du pivot: " + id);
            System.out.println("Valeur du pivot: " + val_pivot);
        }
        for (int j = 0; j < mat.getCol(); j++)
            if (signe(mat.getData(mat.getLine() - 1, j))) {
                has_neg = true;
                break;
            }
        if (has_neg)
            pivot(mat, phase);
    }

    int ligneSortante(Matrix mat, int y, boolean phase) {
        int depth = phase ? mat.getLine() - 2 : mat.getLine() - 1;
        int id = 0;
        while (!(mat.getData(id, y) > 0) && mat.getData(id, depth) >= 0) {
            id++;
        }
        double tmp = mat.getData(id, mat.getCol() - 1) / mat.getData(id, y);
        double tmp_s;
        for (int i = 1; i < depth; i++) {
            if (mat.getData(i, y) > 0) {
                tmp_s = mat.getData(i, mat.getCol() - 1) / mat.getData(i, y);
                if (tmp_s < tmp) {
                    id = i;
                    tmp = tmp_s;
                }
            }
        }
        return id;
    }

    public void printSimplex(Matrix mat, String s) {
        mat.matrixPrint(s);
    }

    boolean signe(Double x) {
        return x < 0;
    }
}
